# Digital IO
IO_INFO = [
    # Right actuator: Inside the forebaffle
    {'name': 'LSR1', 'io': 'IN[1]', 'label': 'Right1Inside'},
    # Right actuator: Outside the forebaffle
    {'name': 'LSR2', 'io': 'IN[2]', 'label': 'Right2Outside'},
    # Right actuator: Spare
    {'name': 'LSR3', 'io': 'IN[3]', 'label': 'Right3NotConnected'},
    # Right actuator: Spare
    {'name': 'LSR4', 'io': 'IN[4]', 'label': 'Right4NotConnected'},
    # Left actuator: Inside the forebaffle
    {'name': 'LSL1', 'io': 'IN[5]', 'label': 'Left1Inside'},
    # Left actuator: Outside the forebaffle
    {'name': 'LSL2', 'io': 'IN[6]', 'label': 'Left2Outside'},
    # Left actuator: Spare
    {'name': 'LSL3', 'io': 'IN[7]', 'label': 'Left3NotConnected'},
    # Left actuator: Spare
    {'name': 'LSL4', 'io': 'IN[8]', 'label': 'Left4NotConnected'},
]
