# Digital IO
IO_INFO = [
    # Right stopper: Inside of the forebaffle interface
    {'name': 'STR1', 'io': 'OUT[1]', 'label': 'Right1Inside'},
    # Right stopper: Outside of the forebaffle interface
    {'name': 'STR2', 'io': 'OUT[2]', 'label': 'Right2Outside'},
    # Left stopper: Inside of the forebaffle interface
    {'name': 'STL1', 'io': 'OUT[3]', 'label': 'Left1Inside'},
    # Left stopper: Outside of the forebaffle interface
    {'name': 'STL2', 'io': 'OUT[4]', 'label': 'Left2Outside'},
]
